==> kite-1.0a4.tar.gz: kite-1.0a4/objs/kite_symtab.c <==
/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "kite_object.h"

/*****************************************************************************
 * Find rightmost element
 ****************************************************************************/
static kite_symtab_t **kite_rightmost_symtab(kite_symtab_t **start) {
    assert(*start != NULL);
    while((*start)->right) start = &(*start)->right;
    return start;
}

/*****************************************************************************
 * Copy a symbol table.
 ****************************************************************************/
void kite_copy_symtab(kite_symtab_t **dest, kite_symtab_t *src) {
    assert(src != NULL);
    
    *dest = malloc(sizeof(kite_symtab_t));
    assert(*dest != NULL);
    
    if (src->left) kite_copy_symtab(&(*dest)->left, src->left);
    else (*dest)->left = NULL;
    if (src->right) kite_copy_symtab(&(*dest)->right, src->right);
    else (*dest)->right = NULL;
    
    (*dest)->name = strdup(src->name);
    (*dest)->global = src->global;
    (*dest)->value = kite_reference_object(src->value);
}

/*****************************************************************************
 * Destroy a symbol table.
 ****************************************************************************/
void kite_destruct_symtab(kite_thread_t *thd, kite_symtab_t **obj, int r)
{
    assert(*obj != NULL);
    
    if ((*obj)->left) kite_destruct_symtab(thd, &(*obj)->left, r);
    if ((*obj)->right) kite_destruct_symtab(thd, &(*obj)->right, r);
    
    free((*obj)->name);
    if (r) kite_dereference_object((*obj)->value);
    free(*obj);
    *obj = NULL;
}

/*****************************************************************************
 * Insert object into symbol table
 ****************************************************************************/
void 
kite_symtab_insert(kite_thread_t *thd, kite_symtab_t **st, kite_symtab_t *obj) {
    kite_symtab_t **pos = st;
    int r;
    
    /* Empty table case */
    if (!*pos) {
        *pos = obj;
        return;
    }
    
    /* Find correct location */
    while(*pos) {
        r = strcmp(obj->name, (*pos)->name);
        if (!r && (*pos)->global == obj->global) {
            /* Replace element */
            obj->left = (*pos)->left;
            obj->right = (*pos)->right;
            kite_dereference_object((*pos)->value);
            free((*pos)->name);
            free(*pos);
            *pos = obj;
            return;
        } else if (r < 0 || (!r && (*pos)->global < obj->global)) {
            pos = &(*pos)->left;
            if (!*pos) {
                kite_symtab_insert(thd, pos, obj);
                return;
            }
        } else {
            pos = &(*pos)->right;
            if (!*pos) {
                kite_symtab_insert(thd, pos, obj);
                return;
            }
        }
    }
}

/*****************************************************************************
 * Remove object from symbol table.
 ****************************************************************************/
void kite_symtab_remove(kite_thread_t *thd, kite_symtab_t **st, char *name) {
    kite_symtab_t **replacement_root, *us = *st, *new_right;
    int r;
    
    assert(*st != NULL);
    r = strcmp((*st)->name, name);
    
    if (!r) {
        replacement_root = kite_rightmost_symtab(&(*st)->left);
        kite_dereference_object(us->value);
        free(us->name);
        
        *st = *replacement_root;
        if (*st) {
            new_right = (*replacement_root)->left;
            (*st)->left = us->left;
            (*st)->right = us->right;
            *replacement_root = new_right;
        }
        
        free(us);
    } else if (r < 0) {
        kite_symtab_remove(thd, &(*st)->left, name);
    } else {
        kite_symtab_remove(thd, &(*st)->right, name);
    }
}

==> kite-1.0.0.tar.gz: kite-1.0.0/objs/kite_symtab.c <==
/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "kite_object.h"

/*****************************************************************************
 * Find rightmost element
 ****************************************************************************/
static kite_symtab_t **kite_rightmost_symtab(kite_symtab_t **start) {
    if (*start != NULL) {
        while((*start)->right) start = &(*start)->right;
        return start;
    } else return NULL;
}

/*****************************************************************************
 * Copy a symbol table's elements.
 ****************************************************************************/
void 
kite_copy_symtab_elements(kite_thread_t *thd, 
                          kite_symtab_t **dest, 
                          kite_symtab_t *src,
                          kite_object_t *p) 
{
    kite_symtab_t *newVal;
#ifdef HAVE_GC_H
    char *tmp;
#endif /* HAVE_GC_H */

    if (src->left) kite_copy_symtab_elements(thd, dest, src->left, p);
#ifndef HAVE_GC_H
    newVal = (kite_symtab_t*)calloc(1, sizeof(kite_symtab_t));
#else
    newVal = (kite_symtab_t*)GC_malloc(sizeof(kite_symtab_t));
#endif
    memcpy(newVal, src, sizeof(kite_symtab_t));
    newVal->left = newVal->right = NULL;
#ifndef HAVE_GC_H
    newVal->name = strdup(newVal->name);
    if (src->docstring) newVal->docstring = strdup(src->docstring);
    else newVal->docstring = NULL;
#else
    tmp = GC_malloc_atomic(strlen(newVal->name) + 1);
    strcpy(tmp, newVal->name);
    newVal->name = tmp;
    if (newVal->docstring) 
    {
        tmp = GC_malloc_atomic(strlen(newVal->docstring) + 1);
        strcpy(tmp, newVal->docstring);
        newVal->docstring = tmp;
    }
#endif /* HAVE_GC_H */
    kite_reference_object(newVal->value);
    newVal->value->parent = p;
    kite_reference_object(p);
    kite_symtab_insert(thd, dest, newVal);
    if (src->right) kite_copy_symtab_elements(thd, dest, src->right, p);
}

/*****************************************************************************
 * Copy a symbol table.
 ****************************************************************************/
void kite_copy_symtab(kite_symtab_t **dest, kite_symtab_t *src) {
    assert(src != NULL);
    
#ifndef HAVE_GC_H
    *dest = (kite_symtab_t*)malloc(sizeof(kite_symtab_t));
#else
    *dest = (kite_symtab_t*)GC_malloc(sizeof(kite_symtab_t));
#endif /* HAVE_GC_H */
    assert(*dest != NULL);
    
    if (src->left) kite_copy_symtab(&(*dest)->left, src->left);
    else (*dest)->left = NULL;
    if (src->right) kite_copy_symtab(&(*dest)->right, src->right);
    else (*dest)->right = NULL;
    
#ifndef HAVE_GC_H
    (*dest)->name = strdup(src->name);
    if (src->docstring) {
        (*dest)->docstring = strdup(src->docstring);
    } else {
        (*dest)->docstring = NULL;
    }
#else
    (*dest)->name = GC_malloc_atomic(strlen(src->name) + 1);
    strcpy((*dest)->name, src->name);
    if (src->docstring) {
        (*dest)->docstring = GC_malloc_atomic(strlen(src->docstring) + 1);
        strcpy((*dest)->docstring, src->docstring);
    } else {
        (*dest)->docstring = NULL;
    }
#endif /* HAVE_GC_H */
    (*dest)->global = src->global;
    (*dest)->value = kite_reference_object(src->value);
}

/*****************************************************************************
 * Destroy a symbol table.
 ****************************************************************************/
void kite_destruct_symtab(kite_thread_t *thd, kite_symtab_t **obj, int r)
{
    assert(*obj != NULL);
    
    if ((*obj)->left) kite_destruct_symtab(thd, &(*obj)->left, r);
    if ((*obj)->right) kite_destruct_symtab(thd, &(*obj)->right, r);
    
#ifndef HAVE_GC_H
    free((*obj)->name);
#endif /* HAVE_GC_H */
    if (r) kite_dereference_object((*obj)->value);
#ifndef HAVE_GC_H
    free((*obj)->docstring);
    free(*obj);
#endif /* HAVE_GC_H */
    *obj = NULL;
}

/*****************************************************************************
 * Insert object into symbol table
 ****************************************************************************/
void 
kite_symtab_insert(kite_thread_t *thd, kite_symtab_t **st, kite_symtab_t *obj) {
    kite_symtab_t **pos = st;
    int r;
    
    /* Empty table case */
    if (!*pos) {
        *pos = obj;
        return;
    }
    
    /* Find correct location */
    while(*pos) {
        r = strcmp(obj->name, (*pos)->name);
        if (!r && (*pos)->global == obj->global) {
            /* Replace element */
            obj->left = (*pos)->left;
            obj->right = (*pos)->right;
            kite_dereference_object((*pos)->value);
#ifndef HAVE_GC_H
            free((*pos)->name);
            free((*pos)->docstring);
            free(*pos);
#endif /* HAVE_GC_H */
            *pos = obj;
            return;
        } else if (r < 0 || (!r && (*pos)->global > obj->global)) {
            pos = &(*pos)->left;
            if (!*pos) {
                kite_symtab_insert(thd, pos, obj);
                return;
            }
        } else {
            pos = &(*pos)->right;
            if (!*pos) {
                kite_symtab_insert(thd, pos, obj);
                return;
            }
        }
    }
}

/*****************************************************************************
 * Remove object from symbol table.
 ****************************************************************************/
void kite_symtab_remove(kite_thread_t *thd, kite_symtab_t **st, char *name, int n) {
    kite_symtab_t **replacement_root, *us = *st, *new_right;
    int r;
    
    assert(*st != NULL);
    r = strcmp(name, (*st)->name);
    
    if (!r && (*st)->global == n) {
        replacement_root = kite_rightmost_symtab(&(*st)->left);

        kite_dereference_object(us->value);
#ifndef HAVE_GC_H
        free(us->name);
        free(us->docstring);
#endif /* HAVE_GC_H */

        if (replacement_root) *st = *replacement_root;
        else *st = NULL;
        if (*st) {
            new_right = (*replacement_root)->left;
            (*st)->left = us->left;
            (*st)->right = us->right;
            *replacement_root = new_right;
        }

#ifndef HAVE_GC_H
        free(us);
#endif /* HAVE_GC_H */
    } else if (r < 0 || (r == 0 && (*st)->global) > n) {
        kite_symtab_remove(thd, &(*st)->left, name, n);
    } else {
        kite_symtab_remove(thd, &(*st)->right, name, n);
    }
}
